/* Test file for Negative Cycle */

#include <sys/time.h>	/* for timing */
#include <sys/resource.h>

#include "defs.h"	/* definition of edge, etc */
#define SEED	19	/* initial seed for random number generator */

extern void generate_graph();
extern edge **negative_cycle();
edge graph_orig[MAX_EDGE], graph[MAX_EDGE];
edge *edge_list[MAX_EDGE];

/* Create a graph with v vertices and e edges and test it.  It has
 * a negative cycle of length cycle unless cycle < 1 in which case
 * there is no negative cycle.  An optional fourth argument
 * is used as a seed for the random number generator
 * input: v (long), e (long), cycle (long), seed (int) [optional]
 * output: 0 on successful termination, 1 otherwise
 * requires: v <= MAX_VERTEX, e <= MAX_EDGE-1, v >= 2, v < e <= v(v-1),
 *           cycle <= v, no user procedure calls srandom() or random().
 */
main(argc,argv)
int argc;
char* argv[];
{
  long i,j, correct, bad, cw, origin, next;
  long e, v, cycle, length;
  edge **result, **check, *cg[MAX_VERTEX];
  struct rusage rusage1, rusage2;
  long time_sec, time_usec;
  int seed;

  /* initialize variables and random number generator */
  if (argc < 3){
    error("Not enough arguments.\n");
  }
  sscanf(argv[1],"%d",&v);
  if (v > MAX_VERTEX){
    error("Number of vertices too large.\n");
  }
  if (v < 2){
    error("Too few vertices.\n");
  }
  sscanf(argv[2],"%d",&e);
  if (e >= MAX_EDGE){
    error("Number of edges too large.\n");
  }
  if (e <= v || e > v * v){
    error("Bad number of edges with respect to number of vertices.\n");
  }
  sscanf(argv[3],"%d",&cycle);
  if (cycle > v){
    error("Cycle too long.\n");
  }
  if (argc <= 4){
    srandom(SEED);
  }
  else{
    sscanf(argv[4],"%d",&seed);
    srandom(seed);
  }

  /* initialize graph */
  generate_graph(v,e,cycle,graph_orig);
  graph_copy(graph,graph_orig,e);	/* make copy to use for test */
  for (i = 0; i < e; i++){
    edge_list[i] = &graph[i];
  }
  edge_list[i] = NULL;
  if (DEBUG){
    print_el(edge_list);
  }

  /* test graph */
  correct = 1;
  /* CALL TO USER FUNCTIION */
  getrusage(RUSAGE_SELF, &rusage1);	/* get time before call */
  result = negative_cycle(edge_list,v);
  getrusage(RUSAGE_SELF, &rusage2);	/* get time after call */
  time_sec = rusage2.ru_utime.tv_sec - rusage1.ru_utime.tv_sec;
  time_usec = rusage2.ru_utime.tv_usec - rusage1.ru_utime.tv_usec;
  if (time_usec < 0){
    time_sec -= 1;
    time_usec += 1000000;
  }
  
  if (result == NULL){		/* no negative cycle found */
    printf("No negative cycle.");
    if (cycle < 1){
      printf("  Correct.\n");
    }
    else{
      printf("  Wrong!\n");
      correct = 0;
    }
  }
  else{				/* negative cycle found */
    printf("Negative cycle found.");
    /* make sure edges are in the graph */
    check = result;
    bad = 0;
    while (*check != NULL){
      if (! in_graph(graph_orig,e,*check)){
	printf("\n     Edge [%5d,%5d](%6d) not in graph.",(*check)->from,
	       (*check)->to,(*check)->weight);
	bad = 1;
      }
      check++;
    }
    if (bad){
      correct=0;
      printf("\n");
    }
    else{
      /* check if this is really a negative cycle */
      /* initialize check array */
      for (j = 0; j < MAX_VERTEX; j++){
	cg[j] = NULL;
      }
      /* store edges according to "from" vertex */
      check = result;
      j = 0;
      while (*check != NULL){
	cg[(*check)->from] = *check;
	check++;
	j++;
      }
      length = j;
      /* check cycle and find weight */
      origin = (*result)->from;
      next = origin;
      cw = 0;
      bad = 0;
      while (j > 0){
	cw += cg[next]->weight;
	next = cg[next]->to;
	if (cg[next] == NULL){
	  printf("\n     Incomplete cycle.  Missing vertex.");
	  bad = 1;
	  correct = 0;
	  break;
	}
	j--;
      }
      if (! bad){
	if (next != origin){
	  printf("\n     Incomplete cycle.  Does not match origin.\n");
	  correct = 0;
	}
	else if (cw >= 0){
	  printf("\n     Weight of cycle not negative: %d.\n",cw);
	  correct = 0;
	}
	else{
	  printf("  Weight = %d.  Length = %d.\n",cw,length);
	}
      }
      else{
	printf("\n");
      }
      if (DEBUG){
	print_el(result);
      }
    }
  }
  if (correct){
    printf("Test is correct.\n");
  }
  else{
    printf("Errors found.\n");
  }
  if (DEBUG){
    printf("Before: %d.%06d\t After: %d.%06d\n", rusage1.ru_utime.tv_sec,
	   rusage1.ru_utime.tv_usec, rusage2.ru_utime.tv_sec,
	   rusage2.ru_utime.tv_usec);
  }
  printf("Approximate time taken = %d.%06d seconds.\n",time_sec, time_usec);
  exit(0);
}

/*********************
 * Utility Functions *
 *********************/

/**********************************************************
 * Feel free to use these utilities in your implementation.
 * However note that function calls are expensive.
 **********************************************************/

/* Copy one edge list to another.
 * input: g1 (array of pointers to edges), g2 (array of pointers to edges)
 * output: number of edges (long)
 * modifies: g1
 * requires: g1 be at least as big as g2; g2 be terminated by NULL
 */
long el_copy(g1,g2)
edge **g1,**g2;
{
  long i=0;

  while(g1[i] = g2[i]){
    i++;
  }
  return(i);
}

/* Copy one edge array to another.
 * input: g1 (array of edges), g2 (array edges), e (number of edges)
 * output: none
 * modifies: g1
 * requires: the size of g1 is at least e
 */
void graph_copy(g1,g2,e)
edge *g1, *g2;
long e;
{
  long i;

  for (i = 0; i < e; i++){
    edge_copy(g1+i,g2+i);
  }
}

/* Swap two edges in an edge list.
 * input: graph (array of pointers to edges), pos1, pos2 [positions of edges
 *       to be swapped] (long)
 * output: none
 * modifies: *graph
 */
void edge_swap(graph,pos1,pos2)
edge **graph;
long pos1, pos2;
{
  edge *temp;	/* temporary variable for swap */

  temp = graph[pos1];
  graph[pos1] = graph[pos2];
  graph[pos2] = temp;
}

/* Swap two edges e1 and e2 in edge array g.
 * Don't confuse with edge_swap, which is for
 * an edge list.
 * input: g (array of edges), e1, e2 (long)
 * output: none
 * modifies: g
 */
void graph_edge_swap(g,e1,e2)
edge *g;
long e1, e2;
{
  edge temp;

  temp.from = g[e2].from;
  temp.to = g[e2].to;
  temp.weight = g[e2].weight;
  g[e2].from = g[e1].from;
  g[e2].to = g[e1].to;
  g[e2].weight = g[e1].weight;
  g[e1].from = temp.from;
  g[e1].to = temp.to;
  g[e1].weight = temp.weight;
}

/* Compares two edges to see if they are equal
 * input: e1, e2 (pointers to edges)
 * output: 1 if equal; 0 otherwise
 * modifies: nothing
 */
long edge_equal(e1, e2)
edge *e1, *e2;
{
  return((e1->from == e2->from) && (e1->to == e2->to) && 
         (e1->weight == e2->weight));
}

/* Copies edge e2 to edge e1
 * input: e1, e2 (pointers to edges)
 * output: none
 * modifies: e1
 */
static void edge_copy(e1, e2)
edge *e1, *e2;
{
  e1->from = e2->from;
  e1->to = e2->to;
  e1->weight = e2->weight;
}

/* Checks to see if an edge is in an edge array.  Basic O(E) algorithm.
 * input: g (array of edges), ne [number of edges in g] (long)
 *        e (pointer to an edge)
 * output: 1 if e is in g; 0 otherwise
 * modifies: nothing
 */
long in_graph(g,ne,e)
edge *g, *e;
long ne;
{
  long i=0;

  while(i < ne){
    if(edge_equal(g+(i++),e)) return(1);
  }
  return(0);
}

/* Simple formatted printed of an edge list.
 * input: g (array of pointers to edges)
 * output: none
 * modifies: nothing
 * requires: g be terminated by NULL
 */
void print_el(g)
edge **g;
{
  long i=0;
  long started = 0;

  while(g[i] != NULL){
    if (i % 3 == 0 && started) printf("\n");
    printf(" [%5d,%5d](%10d)",g[i]->from,g[i]->to,g[i]->weight);
    started = 1;
    i++;
  }
  printf("\n");
}

/* Print an error message and abort.
 * input: s [message to be printed] (string)
 * output: none, but calls exit(1)
 * modifies: destroys everything
 */
void error(s)
char *s;
{
  fprintf(stderr,s);
  exit(1);
}
